@@ -1,19 +1,20 @@
 /* Echo client (async TCP)
  * Copyright (C) 2000-2003  David Helder
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
  *
- * This program is distributed in the hope that it will be useful,
+ * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor,
+ * Boston, MA  02110-1301  USA
  */
 
 
@@ -122,7 +123,7 @@ async_client_sin_iofunc (GIOChannel* iochannel, GIOCondition condition,
     {
       GIOError error;
       gchar buffer[1024];
-      guint bytes_read;
+      gsize bytes_read;
 
       /* Read the data into our buffer */
       error = g_io_channel_read (iochannel, buffer, 
@@ -185,7 +186,7 @@ async_client_in_iofunc (GIOChannel* iochannel, GIOCondition condition,
     {
       GIOError error;
       gchar buffer[1024];
-      guint bytes_read;
+      gsize bytes_read;
 
       /* Read the data into our buffer */
       error = gnet_io_channel_readline (iochannel, buffer, 
@@ -213,7 +214,7 @@ async_client_in_iofunc (GIOChannel* iochannel, GIOCondition condition,
 	{
 	  GIOChannel* sin;
 	  GIOError error;
-	  guint bytes_written;
+	  gsize bytes_written;
 
 	  sin = (GIOChannel*) data;
 	  error = gnet_io_channel_writen (sin, buffer, bytes_read, 
