@@ -1,5 +1,7 @@
 /* GNet - Networking library
  * Copyright (C) 2000  David Helder
+ * Copyright (C) 2003-2004  Andrew Lanoix
+ * Copyright (C) 2007  Tim-Philipp Müller <tim centricular net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -22,27 +24,38 @@
 
 #include <stdio.h>
 #include <memory.h> /* required for windows */
+#include <string.h> /* needed for g_memmove/memmove */
 
+#include "gnet-private.h"
 
 #define IS_CONNECTED(C)  ((C)->socket != NULL)
 #define BUFFER_LEN	 1024
 
+/* FIXME: these macros are horrid, get rid of them */
 #define IS_WATCHING(C, FLAG) (((C)->watch_flags & (FLAG))?TRUE:FALSE)
 
 #define ADD_WATCH(C, FLAG)	do {			\
   if (!IS_WATCHING(C,FLAG)) 	{			\
     (C)->watch_flags |= (FLAG);				\
     if ((C)->iochannel) {				\
-      if ((C)->watch) g_source_remove ((C)->watch);	\
-      (C)->watch = g_io_add_watch ((C)->iochannel, (C)->watch_flags, async_cb, (C)); \
+      if ((C)->watch)					\
+        _gnet_source_remove ((C)->context, (C)->watch);	\
+      (C)->watch = _gnet_io_watch_add_full ((C)->context,\
+          G_PRIORITY_DEFAULT, (C)->iochannel,		\
+          (C)->watch_flags, async_cb, (C), NULL);	\
  }}} while (0)
 
 #define REMOVE_WATCH(C, FLAG)	do {			\
   if (IS_WATCHING(C,FLAG)) 	{			\
     (C)->watch_flags &= ~(FLAG);			\
     if ((C)->iochannel) {				\
-      if ((C)->watch) g_source_remove ((C)->watch);	\
-      (C)->watch = g_io_add_watch ((C)->iochannel, (C)->watch_flags, async_cb, (C)); \
+      if ((C)->watch)					\
+        _gnet_source_remove ((C)->context, (C)->watch);	\
+        (C)->watch = 0;					\
+      if ((C)->watch_flags != 0)			\
+        (C)->watch = _gnet_io_watch_add_full ((C)->context, \
+            G_PRIORITY_DEFAULT, (C)->iochannel,		\
+            (C)->watch_flags, async_cb, (C), NULL);	\
  }}} while (0)
 
 #define UNSET_WATCH (C, FLAG)
@@ -52,7 +65,7 @@ typedef struct _Write
 {
   gchar* 	buffer;
   gint 		length;
-
+  GDestroyNotify buffer_destroy_cb;
 } Write;
 
 
@@ -170,8 +183,8 @@ gnet_conn_new_inetaddr (const GInetAddr* inetaddr,
  *  @user_data: data to pass to @func on callbacks
  *
  *  Creates a #GConn.  The #GConn is created from the @socket.  The
- *  socket is callee owned - do not delete it.  The callback is called
- *  when events occur.
+ *  socket is callee owned - do not delete it (meaning: #GConn will take
+ *  ownership of the socket).  The callback is called when events occur.
  *
  *  Returns: a #GConn.
  *
@@ -198,7 +211,43 @@ gnet_conn_new_socket (GTcpSocket* socket,
 
 }
 
+/**
+ *  gnet_conn_set_main_context
+ *  @conn: a #GConn
+ *  @context: a #GMainContext, or NULL to use the default GLib main context
+ *
+ *  Sets the GLib #GMainContext to use for asynchronous operations. You should
+ *  call this function right after you create @conn. You must not call this
+ *  function after the actual connection process has started or watches have
+ *  been set up (e.g. for reading, writing or errors).
+ *
+ *  You are very unlikely to ever need this function.
+ *
+ *  Returns: TRUE on success, FALSE on failure.
+ *
+ *  Since: 2.0.8
+ **/
+gboolean
+gnet_conn_set_main_context (GConn * conn, GMainContext * context)
+{
+  g_return_val_if_fail (conn != NULL, FALSE);
 
+  /* Must not be called if already connected or connection pending; should
+   * be okay if the socket is connected but no watches set up yet */
+  g_return_val_if_fail (conn->connect_id == 0 && conn->new_id == 0, FALSE);
+  g_return_val_if_fail (conn->watch == 0, FALSE);
+
+  if (conn->context != context) {
+    if (conn->context)
+      g_main_context_unref (conn->context);
+    if (context)
+      conn->context = g_main_context_ref (context);
+    else
+      conn->context = NULL;
+  }
+
+  return TRUE;
+}
 
 /**
  *  gnet_conn_delete
@@ -206,6 +255,7 @@ gnet_conn_new_socket (GTcpSocket* socket,
  *
  *  Deletes a #GConn.
  *
+ *  Deprecated: Use g_conn_unref(), which does the same.
  **/
 void
 gnet_conn_delete (GConn* conn)
@@ -255,8 +305,7 @@ gnet_conn_unref (GConn* conn)
   if (conn->inetaddr)
     gnet_inetaddr_delete (conn->inetaddr);
 
-  if (conn->buffer)
-    g_free (conn->buffer);
+  g_free (conn->buffer);
 
   g_free (conn);
 }
@@ -320,30 +369,27 @@ gnet_conn_set_callback (GConn* conn, GConnFunc func, gpointer user_data)
  *
  **/
 void
-gnet_conn_connect (GConn* conn)
+gnet_conn_connect (GConn * conn)
 {
-  g_return_if_fail (conn);
-  g_return_if_fail (conn->func);
+  g_return_if_fail (conn != NULL);
+  g_return_if_fail (conn->func != NULL);
 
   /* Ignore if connected or connection pending */
-  if (conn->connect_id || conn->new_id || conn->socket)
+  if (conn->connect_id != 0 || conn->new_id != 0 || conn->socket != NULL)
     return;
 
   /* Make asynchronous connection */
-  if (conn->inetaddr)
-    {
-      conn->new_id = 
-	gnet_tcp_socket_new_async (conn->inetaddr, conn_new_cb, conn);
-    }
-  
-  else if (conn->hostname)
-    {
-      conn->connect_id = 
-	gnet_tcp_socket_connect_async (conn->hostname, conn->port, 
-				       conn_connect_cb, conn);
-    }
-  else
-    g_return_if_fail (FALSE);
+  if (conn->inetaddr) {
+    conn->new_id = gnet_tcp_socket_new_async_full (conn->inetaddr,
+        conn_new_cb, conn, (GDestroyNotify) NULL, conn->context,
+        G_PRIORITY_DEFAULT);
+  } else if (conn->hostname) {
+    conn->connect_id = gnet_tcp_socket_connect_async_full (conn->hostname,
+        conn->port, conn_connect_cb, conn, (GDestroyNotify) NULL,
+        conn->context, G_PRIORITY_DEFAULT);
+  } else {
+    g_return_if_reached ();
+  }
 }
 
 
@@ -414,6 +460,14 @@ conn_connect_cb (GTcpSocket* socket,
 }
 
 
+static void
+conn_write_free(Write *write)
+{
+  if (write->buffer_destroy_cb)
+    write->buffer_destroy_cb(write->buffer);
+  g_free (write);
+}
+
 
 /**
  *  gnet_conn_disconnect
@@ -433,7 +487,7 @@ gnet_conn_disconnect (GConn* conn)
 
   if (conn->watch)
     {
-      g_source_remove (conn->watch);
+      _gnet_source_remove (conn->context, conn->watch);
       conn->watch = 0;
     }
   conn->watch_flags = 0;
@@ -464,10 +518,7 @@ gnet_conn_disconnect (GConn* conn)
 
   for (i = conn->write_queue; i != NULL; i = i->next)
     {
-      Write* write = i->data;
-      
-      g_free (write->buffer);
-      g_free (write);
+      conn_write_free(i->data);
     }
   g_list_free (conn->write_queue);
   conn->write_queue = NULL;
@@ -484,13 +535,13 @@ gnet_conn_disconnect (GConn* conn)
   conn->read_eof = FALSE;
   if (conn->process_buffer_timeout)
     {
-      g_source_remove (conn->process_buffer_timeout);
+      _gnet_source_remove (conn->context, conn->process_buffer_timeout);
       conn->process_buffer_timeout = 0;
     }
 
   if (conn->timer)
     {
-      g_source_remove (conn->timer);
+      _gnet_source_remove (conn->context, conn->timer);
       conn->timer = 0;
     }
 }
@@ -525,10 +576,10 @@ async_cb (GIOChannel* iochannel, GIOCondition condition, gpointer data)
   GConn* conn = (GConn*) data;
   GConnEvent event = {GNET_CONN_ERROR, NULL, 0};
   
-  if (condition & (G_IO_ERR | G_IO_HUP | G_IO_NVAL))
-    {
-      ref_internal (conn);
+  ref_internal (conn);
 
+  if (condition & (G_IO_ERR | G_IO_NVAL))
+    {
       /* Upcall ERROR */
       gnet_conn_disconnect (conn);
       if (conn->func)
@@ -540,8 +591,6 @@ async_cb (GIOChannel* iochannel, GIOCondition condition, gpointer data)
 
   if (condition & G_IO_IN)
     {
-      ref_internal (conn);
-
       /* Upcall */
       if (conn->watch_readable)		/* READABLE */
 	{
@@ -564,13 +613,10 @@ async_cb (GIOChannel* iochannel, GIOCondition condition, gpointer data)
 	  unref_internal (conn);
 	  return FALSE;
 	}
-      unref_internal (conn);
     }
 
   if (condition & G_IO_OUT)
     {
-      ref_internal (conn);
-
       /* Upcall */
       if (conn->watch_writable)		/* WRITABLE */
 	{
@@ -593,9 +639,17 @@ async_cb (GIOChannel* iochannel, GIOCondition condition, gpointer data)
 	  unref_internal (conn);
 	  return FALSE;
 	}
-      unref_internal (conn);
     }
 
+  if (condition & G_IO_HUP)
+    {
+      gnet_conn_disconnect (conn);
+      event.type = GNET_CONN_CLOSE;
+      if (conn->func)
+	(conn->func) (conn, &event, conn->user_data);
+    }
+
+  unref_internal (conn);
   /* Assume we want another callback, even though the source may have
      been destroyed though.  The Glib main loop handles this
      properly. */
@@ -719,18 +773,14 @@ conn_check_read_queue (GConn* conn)
      OPTIMIZATION: There may be reads in the queue that cannot be
      processed given the data we have and we would need to read in
      more data.  We watch IO_IN in this case.  */
-  if ((conn->bytes_read && bytes_processable(conn) > 0) || conn->read_eof)
-    {
-      conn->process_buffer_timeout = 
-	g_timeout_add (0, process_read_buffer_cb, conn);
-    }
-
+  if ((conn->bytes_read && bytes_processable(conn) > 0) || conn->read_eof) {
+    conn->process_buffer_timeout = _gnet_timeout_add_full (conn->context,
+        G_PRIORITY_DEFAULT, 0, process_read_buffer_cb, conn, NULL);
+  } else {
   /* Otherwise, if there is no read watch, set one, so we can read
-     more bytes */
-  else
-    {
-      ADD_WATCH(conn, G_IO_IN);
-    }
+   * more bytes */
+    ADD_WATCH(conn, G_IO_IN);
+  }
 }
 
 
@@ -740,7 +790,7 @@ conn_read_async_cb (GConn* conn)
   guint  bytes_to_read;
   gchar* buffer_start;
   GIOError error;
-  guint  bytes_read;
+  gsize  bytes_read;
   
   /* Resize the buffer if it's full. */
   if (conn->length == conn->bytes_read)
@@ -922,7 +972,7 @@ bytes_processable (GConn* conn)
       /* Read line */
     case -1:		
       {
-	gint i;
+	guint i;
 
 	/* Look for \n, \r, or \r\n */
 	for (i = 0; i < conn->bytes_read; ++i)
@@ -990,7 +1040,7 @@ process_read_buffer (GConn* conn)
       /* Read line */
     case -1:		
       {
-	gint i;
+	guint i;
 
 	/* Look for \n, \r, or \r\n */
 	for (i = 0; i < conn->bytes_read; ++i)
@@ -1083,6 +1133,43 @@ process_read_buffer (GConn* conn)
 
 
 
+/**
+ *  gnet_conn_write_direct
+ *  @conn: a #GConn
+ *  @buffer: buffer to write from
+ *  @length: length of @buffer
+ *  @buffer_destroy_cb: function to call when buffer is no longer needed
+ *
+ *  Sets up an asynchronous write to @conn from @buffer.  The buffer is
+ *  created by the caller and will not be copied.  The caller needs to
+ *  make sure the buffer stays valid until @buffer_destroy_cb is called.
+ *  This function can be called again before the asynchronous write
+ *  completes.  @buffer_destroy_cb may be %NULL.
+ *
+ **/
+void
+gnet_conn_write_direct (GConn* conn, gchar* buffer, gint length,
+			GDestroyNotify buffer_destroy_cb)
+{
+  Write* write;
+
+  g_return_if_fail (conn != NULL);
+  g_return_if_fail (buffer != NULL);
+  g_return_if_fail (length >= 0);
+
+  if (length == 0)
+    return;
+
+  /* Add to queue */
+  write = g_new0 (Write, 1);
+  write->buffer = buffer;
+  write->length = length;
+  write->buffer_destroy_cb = buffer_destroy_cb;
+  conn->write_queue = g_list_append (conn->write_queue, write);
+
+  conn_check_write_queue (conn);
+}
+
 /**
  *  gnet_conn_write
  *  @conn: a #GConn
@@ -1097,18 +1184,17 @@ process_read_buffer (GConn* conn)
 void
 gnet_conn_write (GConn* conn, gchar* buffer, gint length)
 {
-  Write* write;
+  gchar *copy;
 
-  g_return_if_fail (conn);
-  g_return_if_fail (conn->func);
+  g_return_if_fail (conn != NULL);
+  g_return_if_fail (buffer != NULL);
+  g_return_if_fail (length >= 0);
 
-  /* Add to queue */
-  write = g_new0 (Write, 1);
-  write->buffer = g_memdup (buffer, length);
-  write->length = length;
-  conn->write_queue = g_list_append (conn->write_queue, write);
+  if (length == 0)
+    return;
 
-  conn_check_write_queue (conn);
+  copy = g_memdup (buffer, length);
+  gnet_conn_write_direct (conn, copy, length, (GDestroyNotify) g_free);
 }
 
 
@@ -1135,7 +1221,7 @@ conn_write_async_cb (GConn* conn)
   GIOError   error;
   guint      bytes_to_write;
   gchar*     buffer_start;
-  guint      bytes_written;
+  gsize      bytes_written;
   GConnEvent event = {GNET_CONN_ERROR, NULL, 0};
 
   write = (Write*) conn->write_queue->data;
@@ -1167,8 +1253,7 @@ conn_write_async_cb (GConn* conn)
     {
       /* Remove and delete the queued write */
       conn->write_queue = g_list_remove (conn->write_queue, write);
-      g_free (write->buffer);
-      g_free (write);
+      conn_write_free(write);
 
       conn->bytes_written = 0;
 
@@ -1191,7 +1276,7 @@ conn_write_async_cb (GConn* conn)
 /* **************************************** */
 
 /**
- * gnet_conn_set_watch_readable
+ * gnet_conn_set_watch_error
  * @conn: a #GConn
  * @enable: enable the %GNET_CONN_READABLE event?
  *
@@ -1279,19 +1364,18 @@ gnet_conn_set_watch_writable (GConn* conn, gboolean enable)
 void
 gnet_conn_timeout (GConn* conn, guint timeout)
 {
-  g_return_if_fail (conn);
+  g_return_if_fail (conn != NULL);
 
-  if (conn->timer)
-    {
-      g_source_remove (conn->timer);
-      conn->timer = 0;
-    }
+  if (conn->timer) {
+    _gnet_source_remove (conn->context, conn->timer);
+    conn->timer = 0;
+  }
 
-  if (timeout)
-    {
-      g_return_if_fail (conn->func);
-      conn->timer = g_timeout_add (timeout, conn_timeout_cb, conn);
-    }
+  if (timeout) {
+    g_return_if_fail (conn->func != NULL);
+    conn->timer = _gnet_timeout_add_full (conn->context, G_PRIORITY_DEFAULT,
+        timeout, conn_timeout_cb, conn, (GDestroyNotify) NULL);
+  }
 }
 
 
