@@ -1,6 +1,7 @@
 /* GNet - Networking library
  * Copyright (C) 2000  David Helder
  * Copyright (C) 2000-2003  Andrew Lanoix
+ * Copyright (C) 2007 Tim-Philipp Müller <tim centricular net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -37,10 +38,6 @@
 #include <sys/param.h>
 #endif
 
-#ifdef HAVE_LIBPTHREAD
-#include <pthread.h>
-#endif
-
 #include <glib.h>
 #include "gnet.h"
 #include "gnetconfig.h"
@@ -86,36 +83,38 @@ struct sockaddr_storage {
 #define SOCKET gint
 #endif
 
-#ifndef socklen_t
-#  ifdef GNET_APPLE_DARWIN
-#    define socklen_t int	/* socklen_t is int in Darwin */
-#  else
-#    define socklen_t size_t	/* it's size_t on all other Unixes */
-#  endif
-#endif
-
 #define GNET_CLOSE_SOCKET(SOCKFD) close(SOCKFD)
 
-/* Use gnet_private_io_channel_new() to create iochannels */
+/* Use _gnet_io_channel_new() to create iochannels */
 #define GNET_SOCKET_IO_CHANNEL_NEW(SOCKFD) g_io_channel_unix_new(SOCKFD)
 
 #define GNET_IS_SOCKET_VALID(S) ((S) >= 0)
+#define GNET_INVALID_SOCKET (-1)
+
+#define STATUS_IS_SOCKET_ERROR(status) ((status) < 0)
+
+#define ERROR_IS_CONNECT_IN_PROGRESS() (errno == EINPROGRESS)
 
 #else	/*********** Windows specific ***********/
 
-#include <windows.h>
-#include <winbase.h>
-#include <winuser.h>
+#include <winsock2.h>
 #include <ws2tcpip.h>
 
-#define socklen_t gint32
+#ifndef socklen_t
+#  define socklen_t gint32
+#endif
 #define in_addr_t guint32
 
+#define STATUS_IS_SOCKET_ERROR(status) ((status) == SOCKET_ERROR)
+
+#define ERROR_IS_CONNECT_IN_PROGRESS() (WSAGetLastError () == WSAEWOULDBLOCK)
+
 #define GNET_IS_SOCKET_VALID(S) ((S) != INVALID_SOCKET)
+#define GNET_INVALID_SOCKET (INVALID_SOCKET)
 
 #define GNET_CLOSE_SOCKET(SOCKFD) closesocket(SOCKFD)
 
-/* Use gnet_private_io_channel_new() to create iochannels */
+/* Use _gnet_io_channel_new() to create iochannels */
 #define GNET_SOCKET_IO_CHANNEL_NEW(SOCKFD) g_io_channel_win32_new_socket(SOCKFD)
 
 #endif	/*********** End Windows specific ***********/
@@ -132,11 +131,14 @@ struct sockaddr_storage {
 #define IN_LOOPBACKNET 127
 #endif
 
+#ifndef GNET_SOCKADDR_FAMILY_FIELD_NAME
+# define GNET_SOCKADDR_FAMILY_FIELD_NAME ss_family
+#endif
 
-#define GNET_SOCKADDR_IN(s)    	(*((struct sockaddr_in*) &s))
-#define GNET_SOCKADDR_SA(s)	(*((struct sockaddr*) &s))
-#define GNET_SOCKADDR_SA4(s)	(*((struct sockaddr_in*) &s))
-#define GNET_SOCKADDR_FAMILY(s) ((s).ss_family)
+#define GNET_SOCKADDR_IN(s)     (*((struct sockaddr_in*) &s))
+#define GNET_SOCKADDR_SA(s)     (*((struct sockaddr*) &s))
+#define GNET_SOCKADDR_SA4(s)    (*((struct sockaddr_in*) &s))
+#define GNET_SOCKADDR_FAMILY(s) ((s).GNET_SOCKADDR_FAMILY_FIELD_NAME)
 
 #ifdef HAVE_IPV6
 
@@ -149,27 +151,27 @@ struct sockaddr_storage {
 #endif
 
 #define GNET_SOCKADDR_SA6(s)	(*((struct sockaddr_in6*) &s))
-#define GNET_SOCKADDR_ADDRP(s)	(((s).ss_family == AF_INET)?\
+#define GNET_SOCKADDR_ADDRP(s)	((GNET_SOCKADDR_FAMILY((s)) == AF_INET)?\
                                   (void*)&((struct sockaddr_in*)&s)->sin_addr:\
                                   (void*)&((struct sockaddr_in6*)&s)->sin6_addr)
-#define GNET_SOCKADDR_ADDR32(s,n)(((s).ss_family == AF_INET)?\
+#define GNET_SOCKADDR_ADDR32(s,n)((GNET_SOCKADDR_FAMILY((s)) == AF_INET)?\
                                   ((struct sockaddr_in*)&s)->sin_addr.s_addr:\
                                   *(guint32*)&((struct sockaddr_in6*)&s)->sin6_addr.s6_addr[(n)*4])
-#define GNET_SOCKADDR_ADDR32_SET(s,n,a) if ((s).ss_family == AF_INET) \
+#define GNET_SOCKADDR_ADDR32_SET(s,n,a) if (GNET_SOCKADDR_FAMILY((s)) == AF_INET) \
                                           ((struct sockaddr_in*)&s)->sin_addr.s_addr = a; \
                                         else \
                                           *(guint32*)&((struct sockaddr_in6*)&s)->sin6_addr.s6_addr[(n)*4] = a;
-#define GNET_SOCKADDR_ADDRLEN(s) (((s).ss_family == AF_INET)?\
+#define GNET_SOCKADDR_ADDRLEN(s) ((GNET_SOCKADDR_FAMILY((s)) == AF_INET)?\
 				 sizeof(struct in_addr):\
 				 sizeof(struct in6_addr))
-#define GNET_SOCKADDR_PORT(s)	(((s).ss_family == AF_INET)?\
+#define GNET_SOCKADDR_PORT(s)	((GNET_SOCKADDR_FAMILY((s)) == AF_INET)?\
                                   ((struct sockaddr_in*)&s)->sin_port:\
                                   ((struct sockaddr_in6*)&s)->sin6_port)
-#define GNET_SOCKADDR_PORT_SET(s, p)	if ((s).ss_family == AF_INET)\
+#define GNET_SOCKADDR_PORT_SET(s, p)	if (GNET_SOCKADDR_FAMILY((s)) == AF_INET)\
                                           ((struct sockaddr_in*)&(s))->sin_port = p;\
                                         else \
                                           ((struct sockaddr_in6*)&(s))->sin6_port = p;
-#define GNET_SOCKADDR_LEN(s)	(((s).ss_family == AF_INET)?\
+#define GNET_SOCKADDR_LEN(s)	((GNET_SOCKADDR_FAMILY((s)) == AF_INET)?\
                                   sizeof(struct sockaddr_in):\
                                   sizeof(struct sockaddr_in6))
 #else /* NO IPV6 */
@@ -206,9 +208,7 @@ struct sockaddr_storage {
 
 #define GNET_ANY_IO_CONDITION   (G_IO_IN|G_IO_OUT|G_IO_PRI|G_IO_ERR|G_IO_HUP|G_IO_NVAL)
 
-#ifdef __cplusplus
-extern "C" {
-#endif /* __cplusplus */
+G_BEGIN_DECLS
 
 
 /*
@@ -220,28 +220,27 @@ extern "C" {
 
 */
 
-
+#define GNET_UDP_SOCKET_TYPE_COOKIE    71254329
+#define GNET_MCAST_SOCKET_TYPE_COOKIE  49712423
 
 struct _GUdpSocket
 {
+  guint type;
   SOCKET sockfd;
-  guint ref_count;
+  gint ref_count;
   GIOChannel* iochannel;
   struct sockaddr_storage sa;
 };
 
 struct _GMcastSocket
 {
-  SOCKET sockfd;
-  guint ref_count;
-  GIOChannel* iochannel;
-  struct sockaddr_storage sa;
+  GUdpSocket udpsocket;
 };
 
 struct _GTcpSocket
 {
   SOCKET sockfd;
-  guint ref_count;
+  gint ref_count; /* ATOMIC */
   GIOChannel* iochannel;
   struct sockaddr_storage sa;
   /* sa is remote host for clients, local host for servers */
@@ -254,167 +253,15 @@ struct _GTcpSocket
 struct _GInetAddr
 {
   gchar* name;
-  guint ref_count;
+  gint ref_count; /* ATOMIC */
   struct sockaddr_storage sa;
 };
 
-
-#ifndef GNET_WIN32  /*********** Unix specific ***********/
-
-struct _GUnixSocket
-{
-  gint sockfd;
-  guint ref_count;
-  GIOChannel *iochannel;
-  struct sockaddr_un sa;
-
-  gboolean server;
-};
-
-#endif	/*********** End Unix specific ***********/
-
-
-
-/* **************************************** */
-/* Async functions			*/
-
-gboolean gnet_inetaddr_new_list_async_cb (GIOChannel* iochannel, 
-					  GIOCondition condition, 
-					  gpointer data);
-
-
-typedef struct _GInetAddrNewListState 
-{
-  GList*	ias;
-  gint		port;
-  GInetAddrNewListAsyncFunc func;
-  gpointer 	data;
-
-#ifndef GNET_WIN32		/* UNIX */
-  gboolean 	in_callback;
-#ifdef HAVE_LIBPTHREAD		/* UNIX pthread	*/
-  pthread_mutex_t mutex;
-  gboolean	is_cancelled;
-  gboolean	lookup_failed;
-  guint 	source;
-
-#else			       	/* UNIX process	*/
-  int 		fd;
-  pid_t 	pid;
-  GIOChannel* 	iochannel;
-  guint 	watch;
-  int 		len;
-  guchar 	buffer[256];
-
-#endif
-#else				/* Windows */
-  gboolean 	in_callback;
-  CRITICAL_SECTION mutex;
-  gboolean	is_cancelled;
-  gboolean	lookup_failed;
-  guint 	source;
-#endif
-
-} GInetAddrNewListState;
-
-
-
-typedef struct _GInetAddrNewState 
-{
-  GInetAddrNewListAsyncID	list_id;
-  GInetAddrNewAsyncFunc 	func;
-  gpointer 			data;
-  gboolean			in_callback;
-
-} GInetAddrNewState;
-
-
-gboolean gnet_inetaddr_get_name_async_cb (GIOChannel* iochannel, 
-					  GIOCondition condition, 
-					  gpointer data);
-
-typedef struct _GInetAddrReverseAsyncState 
-{
-  GInetAddr* ia;
-  GInetAddrGetNameAsyncFunc func;
-  gpointer data;
-  gboolean in_callback;
-#ifndef GNET_WIN32		/* UNIX 	*/
-#ifdef HAVE_LIBPTHREAD		/* UNIX pthread	*/
-  pthread_mutex_t mutex;
-  gboolean	is_cancelled;
-  gchar*	name;
-  guint 	source;
-#else				/* UNIX process	*/
-  int 		fd;
-  pid_t 	pid;
-  guint 	watch;
-  GIOChannel* 	iochannel;
-#endif
-  guchar	buffer[256 + 1];/* Names can only be 256 characters? */
-  int 		len;
-#else				/* WINDOWS */
-  int WSAhandle;
-  char hostentBuffer[MAXGETHOSTSTRUCT];
-  int errorcode;
-
-  CRITICAL_SECTION mutex;
-  gboolean	is_cancelled;
-  gchar*	name;
-  guint 	source;
-
-#endif
-
-} GInetAddrReverseAsyncState;
-
-
-gboolean gnet_tcp_socket_new_async_cb (GIOChannel* iochannel, 
-				       GIOCondition condition, 
-				       gpointer data);
-
-typedef struct _GTcpSocketAsyncState 
-{
-  GTcpSocket* 		 socket;
-  GTcpSocketNewAsyncFunc func;
-  gpointer 		 data;
-  gint 			 flags;
-  GIOChannel* 		 iochannel;
-  guint 		 connect_watch;
-#ifdef GNET_WIN32
-  gint 			 errorcode;
-#endif
-} GTcpSocketAsyncState;
-
-void gnet_tcp_socket_connect_inetaddr_cb (GList* ia_list, gpointer data);
-
-void gnet_tcp_socket_connect_tcp_cb(GTcpSocket* socket, gpointer data);
-
-typedef struct _GTcpSocketConnectState 
-{
-  GList* ia_list;
-  GList* ia_next;
-
-  gpointer inetaddr_id;
-  gpointer tcp_id;
-
-  gboolean in_callback;
-
-  GTcpSocketConnectAsyncFunc func;
-  gpointer data;
-
-} GTcpSocketConnectState;
-
-
 /* **************************************** 	*/
 /* More Windows specific stuff 			*/
 
 #ifdef GNET_WIN32
 
-extern HWND  gnet_hWnd; 
-extern GHashTable *gnet_hash;
-#define IA_NEW_MSG 100		/* gnet_inetaddr_new_async */
-#define GET_NAME_MSG 101	/* gnet_inetaddr_get_name_async */
-
 /* Name-mangled IPv6 structures for primarily Mingw and a few VC .NET 2002 */
 #ifndef MAX_ADAPTER_ADDRESS_LENGTH
 #define MAX_ADAPTER_ADDRESS_LENGTH 8
@@ -653,15 +500,39 @@ PFN_GETADAPTERSADDRESSES pfn_getaddaptersaddresses;
 
 /* Private/Experimental functions */
 
-GIOChannel* gnet_private_io_channel_new (SOCKET sockfd);
+GIOChannel* _gnet_io_channel_new (SOCKET sockfd);
 
-SOCKET gnet_private_create_listen_socket (int type, const GInetAddr* iface, int port, struct sockaddr_storage* sa);
+SOCKET _gnet_create_listen_socket (int type, const GInetAddr* iface, int port, struct sockaddr_storage* sa);
 
 int gnet_initialize_windows_sockets(void);
 void gnet_uninitialize_windows_sockets(void);
 
-#ifdef __cplusplus
-}
-#endif				/* __cplusplus */
+/* Private utility functions */
+
+guint   _gnet_idle_add_full     (GMainContext  * context,
+                                 gint            priority,
+                                 GSourceFunc     function,
+                                 gpointer        data,
+                                 GDestroyNotify  notify);
+
+guint   _gnet_timeout_add_full  (GMainContext  * context,
+                                 gint            priority,
+                                 guint           interval,
+                                 GSourceFunc     function,
+                                 gpointer        data,
+                                 GDestroyNotify  notify);
+
+guint   _gnet_io_watch_add_full (GMainContext  * context,
+                                 gint            priority,
+                                 GIOChannel    * channel,
+                                 GIOCondition    condition,
+                                 GIOFunc         function,
+                                 gpointer        data,
+                                 GDestroyNotify  notify);
+
+/* will do nothing if source_id is 0 (unlike g_source_remove()) */
+void    _gnet_source_remove     (GMainContext * context, guint source_id);
+
+G_END_DECLS
 
 #endif /* _GNET_PRIVATE_H */
