@@ -29,10 +29,12 @@ const guint gnet_interface_age = GNET_INTERFACE_AGE;
 const guint gnet_binary_age = GNET_BINARY_AGE;
 
 
-#ifdef HAVE_IPV6
+#if !defined(GNET_WIN32) && defined(HAVE_IPV6)
+#ifndef GNET_WIN32
 static gboolean ipv6_detect_envvar (void);
 static gboolean ipv6_detect_iface (void);
 #endif
+#endif
 
 #ifdef GNET_WIN32
 void gnet_win32_at_exit( void ); /* Here to keep Visual Studio happy. */
@@ -60,13 +62,16 @@ void gnet_win32_at_exit()
 void
 gnet_init (void)
 {
-#ifdef G_THREADS_ENABLED
-#ifndef GNET_WIN32 
+  static gboolean  been_here; /* FALSE */
+
+  if (been_here)
+    return;
+
+  /* TODO: use atomic ops once we require Glib-2.4 */
+  been_here = TRUE;
+
   if (!g_thread_supported ()) 
     g_thread_init (NULL);
-#endif
-#endif /* G_THREADS_ENABLED */
-
 
 #ifndef GNET_WIN32
   /* Auto-detect IPv6 policy.  Set it to IPv4 if auto-detection fails. */
@@ -115,12 +120,14 @@ gnet_init (void)
 	gnet_ipv6_set_policy (policy);
 	gnet_initialize_windows_sockets();
 
+	/* FIXME: a library should really not set up atexit handlers ... */
 	atexit(gnet_win32_at_exit);
 #endif
 }
 
 
-#ifdef HAVE_IPV6
+#if !defined(GNET_WIN32) && defined(HAVE_IPV6)
+#ifndef GNET_WIN32
 /* 
 
    Try to get policy based on environment variables.  We look in the
@@ -214,3 +221,4 @@ ipv6_detect_iface (void)
   return TRUE;
 }
 #endif
+#endif
