@@ -21,8 +21,8 @@
 #include <string.h>
 
 
-static gsize strlenn(char* str, gsize n);
-static void flipmemcpy(char* dst, char* src, gsize n);
+static inline gsize strlenn(const char* str, gsize n);
+static inline void flipmemcpy(char* dst, const char* src, gsize n);
 
 
 #define MEMCPY(D,S,N)				\
@@ -78,7 +78,7 @@ static void flipmemcpy(char* dst, char* src, gsize n);
 
 */
 static gsize
-strlenn(char* str, gsize n)
+strlenn(const char* str, gsize n)
 {
   gsize len = 0;
 
@@ -89,12 +89,12 @@ strlenn(char* str, gsize n)
 
 
 static void 
-flipmemcpy(char* dst, char* src, gsize n)
+flipmemcpy(char* dst, const char* src, gsize n)
 {
-  int nn = n;
-
-  for (; n; --n)
-    *dst++ = src[nn-n];
+  while (n > 0) {
+    *dst++ = src[n-1];
+    --n;
+  }
 }
 
 
@@ -230,7 +230,7 @@ flipmemcpy(char* dst, char* src, gsize n)
  *  l/L is a signed/unsigned long.
  *
  *  f/D is a float/double (always native order/size).
- *  
+ *
  *  v is a void pointer (always native size).
  *
  *  s is a zero-terminated string.
@@ -833,7 +833,7 @@ gnet_vpack (const gchar* format, gchar* buffer, const gint length, va_list args)
  * 
  **/
 gint 
-gnet_unpack (const gchar* format, gchar* buffer, gint length, ...)
+gnet_unpack (const gchar * format, const gchar * buffer, gint length, ...)
 {
   va_list args;
   gint rv;
@@ -860,7 +860,8 @@ gnet_unpack (const gchar* format, gchar* buffer, gint length, ...)
  *
  **/
 gint 
-gnet_vunpack (const gchar* format, gchar* buffer, gint length, va_list args)
+gnet_vunpack (const gchar * format, const gchar * buffer, gint length,
+    va_list args)
 {
   gint n = 0;
   gchar* p = (gchar*) format;
