@@ -433,7 +433,7 @@ SHAFinal( char *key, SHA_CTX *shaInfo )
 struct _GSHA
 {
   SHA_CTX	ctx;
-  gchar 	digest[GNET_SHA_HASH_LENGTH];
+  guint8	digest[GNET_SHA_HASH_LENGTH];
 };
 
 
@@ -454,7 +454,7 @@ gnet_sha_new (const gchar* buffer, guint length)
 
   sha = g_new0 (GSHA, 1);
   SHAInit (&sha->ctx);
-  SHAUpdate (&sha->ctx, buffer, length);
+  SHAUpdate (&sha->ctx, (const guchar*) buffer, length);
   SHAFinal ((gpointer) &sha->digest, &sha->ctx);
 
   return sha;
@@ -561,8 +561,7 @@ gnet_sha_clone (const GSHA* sha)
 void
 gnet_sha_delete (GSHA* sha)
 {
-  if (sha)
-    g_free (sha);
+  g_free (sha);
 }
 
 
@@ -606,7 +605,7 @@ gnet_sha_update (GSHA* sha, const gchar* buffer, guint length)
 {
   g_return_if_fail (sha);
 
-  SHAUpdate (&sha->ctx, buffer, length);
+  SHAUpdate (&sha->ctx, (const guchar*) buffer, length);
 }
 
 
@@ -649,10 +648,7 @@ gnet_sha_get_digest (const GSHA* sha)
 }
 
 
-static gchar bits2hex[16] = { '0', '1', '2', '3', 
-			      '4', '5', '6', '7',
-			      '8', '9', 'a', 'b',
-			      'c', 'd', 'e', 'f' };
+static const gchar bits2hex[] = "0123456789abcdef";
 
 /**
  *  gnet_sha_get_string
