@@ -1,6 +1,7 @@
 /* GNet - Networking library
- * Copyright (C) 2000-3  David Helder
- * Copyright (C) 2000  Andrew Lanoix
+ * Copyright (C) 2000-2003  David Helder
+ * Copyright (C) 2000-2004  Andrew Lanoix
+ * Copyright (C) 2007       Tim-Philipp Müller <tim centricular net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -22,6 +23,45 @@
 #include "socks-private.h"
 #include "tcp.h"
 
+static gboolean gnet_tcp_socket_new_async_cb (GIOChannel * iochannel,
+    GIOCondition condition, gpointer data);
+static void gnet_tcp_socket_connect_inetaddr_cb (GList * ia_list, gpointer data);
+static void gnet_tcp_socket_connect_tcp_cb (GTcpSocket * socket, gpointer data);
+
+typedef struct _GTcpSocketAsyncState
+{
+  GTcpSocket             * socket;
+  GTcpSocketNewAsyncFunc   func;
+  gpointer                 data;
+  GDestroyNotify           notify;
+  gint                     flags;
+  GIOChannel             * iochannel;
+  guint                    connect_watch;
+#ifdef GNET_WIN32
+  gint                     errorcode;
+#endif
+  GMainContext           * context; /* we hold a ref */
+  gint                     priority;
+} GTcpSocketAsyncState;
+
+typedef struct _GTcpSocketConnectState
+{
+  GList * ia_list;
+  GList * ia_next;
+
+  GInetAddrNewListAsyncID  inetaddr_id;
+  GTcpSocketNewAsyncID     tcp_id;
+
+  gboolean in_callback;
+
+  GTcpSocketConnectAsyncFunc func;
+  gpointer                   data;
+  GDestroyNotify             notify;
+
+  GMainContext             * context; /* we hold a ref */
+  gint                       priority;
+} GTcpSocketConnectState;
+
 /**
  *  gnet_tcp_socket_connect
  *  @hostname: host name
@@ -63,6 +103,67 @@ gnet_tcp_socket_connect (const gchar* hostname, gint port)
 }
 
 
+/**
+ *  gnet_tcp_socket_connect_async_full
+ *  @hostname: host name
+ *  @port: port
+ *  @func: callback function
+ *  @data: data to pass to @func on callback
+ *  @notify: function to call to free @data, or NULL
+ *  @context: the #GMainContext to use for notifications, or NULL for the
+ *      default GLib main context.  If in doubt, pass NULL.
+ *  @priority: the priority with which to schedule notifications in the
+ *      main context, e.g. #G_PRIORITY_DEFAULT or #G_PRIORITY_HIGH.
+ *
+ *  Asynchronously creates a #GTcpSocket and connects to
+ *  @hostname:@port.  The callback is called when the connection is
+ *  made or an error occurs.  The callback will not be called during
+ *  the call to this function.
+ *
+ *  Returns: the ID of the connection; NULL on failure.  The ID can be
+ *  used with gnet_tcp_socket_connect_async_cancel() to cancel the
+ *  connection.
+ *
+ *  Since: 2.0.8
+ **/
+GTcpSocketConnectAsyncID
+gnet_tcp_socket_connect_async_full (const gchar * hostname, gint port, 
+    GTcpSocketConnectAsyncFunc func, gpointer data, GDestroyNotify notify,
+    GMainContext * context, gint priority)
+{
+  GTcpSocketConnectState* state;
+
+  g_return_val_if_fail (hostname != NULL, NULL);
+  g_return_val_if_fail (func != NULL, NULL);
+
+  if (context == NULL)
+    context = g_main_context_default ();
+
+  state = g_new0 (GTcpSocketConnectState, 1);
+
+  state->func = func;
+  state->data = data;
+  state->notify = notify;
+  state->context = g_main_context_ref (context);
+  state->priority = priority;
+
+  state->inetaddr_id = gnet_inetaddr_new_list_async_full (hostname, port,
+      gnet_tcp_socket_connect_inetaddr_cb, state, (GDestroyNotify) NULL,
+      state->context, priority);
+
+  /* On failure, gnet_inetaddr_new_list_async_full() returns NULL.  It will
+   * not call the callback before it returns. */
+  if (state->inetaddr_id == NULL) {
+    if (state->notify)
+      state->notify (state->data);
+    g_main_context_unref (state->context);
+    g_free (state);
+    return NULL;
+  }
+
+  return state;
+}
+
 /**
  *  gnet_tcp_socket_connect_async
  *  @hostname: host name
@@ -81,61 +182,43 @@ gnet_tcp_socket_connect (const gchar* hostname, gint port)
  *
  **/
 GTcpSocketConnectAsyncID
-gnet_tcp_socket_connect_async (const gchar* hostname, gint port, 
-			       GTcpSocketConnectAsyncFunc func, 
-			       gpointer data)
+gnet_tcp_socket_connect_async (const gchar * hostname, gint port, 
+    GTcpSocketConnectAsyncFunc func, gpointer data)
 {
-  GTcpSocketConnectState* state;
+  GTcpSocketConnectAsyncID async_id;
 
-  g_return_val_if_fail(hostname != NULL, NULL);
-  g_return_val_if_fail(func != NULL, NULL);
+  g_return_val_if_fail (hostname != NULL, NULL);
+  g_return_val_if_fail (func != NULL, NULL);
 
-  state = g_new0(GTcpSocketConnectState, 1);
-  state->func = func;
-  state->data = data;
+  async_id = gnet_tcp_socket_connect_async_full (hostname, port, func, data,
+      (GDestroyNotify) NULL, NULL, G_PRIORITY_DEFAULT);
 
-  state->inetaddr_id = 
-    gnet_inetaddr_new_list_async (hostname, port,  
-				  gnet_tcp_socket_connect_inetaddr_cb, 
-				  state);
-
-  /* On failure, gnet_inetaddr_new_list_async() returns NULL.  It will
-     not call the callback before it returns. */
-  if (state->inetaddr_id == NULL)
-    {
-      g_free (state);
-      return NULL;
-    }
-
-  return state;
+  return async_id;
 }
 
-
-
-void
+static void
 gnet_tcp_socket_connect_inetaddr_cb (GList* ia_list, gpointer data)
 {
   GTcpSocketConnectState* state = (GTcpSocketConnectState*) data;
 
   if (ia_list != NULL) /* Success */
     {
-      g_assert (ia_list);
-
       state->inetaddr_id = NULL;
       state->ia_list = ia_list;
       state->ia_next = ia_list;
 
       while (state->ia_next != NULL)
 	{
+	  GTcpSocketNewAsyncID tcp_id;
 	  GInetAddr* ia;
-	  gpointer tcp_id;
 
 	  ia = (GInetAddr*) state->ia_next->data;
 	  state->ia_next = state->ia_next->next;
 
-	  tcp_id = gnet_tcp_socket_new_async (ia, 
-					      gnet_tcp_socket_connect_tcp_cb, 
-					      state);
+	  tcp_id = gnet_tcp_socket_new_async_full (ia,
+              gnet_tcp_socket_connect_tcp_cb, state, (GDestroyNotify) NULL,
+              state->context, state->priority);
+
 	  if (tcp_id)	/* Success */
 	    {
 	      state->tcp_id = tcp_id;
@@ -150,6 +233,7 @@ gnet_tcp_socket_connect_inetaddr_cb (GList* ia_list, gpointer data)
 		     state->data);
       state->in_callback = FALSE;
 
+      /* FIXME: don't abuse _cancel() as free function for the state */
       gnet_tcp_socket_connect_async_cancel (state);
     }
   else /* Failure */
@@ -159,12 +243,13 @@ gnet_tcp_socket_connect_inetaddr_cb (GList* ia_list, gpointer data)
 		     state->data);
       state->in_callback = FALSE;
 
+      /* FIXME: don't abuse _cancel() as free function for the state */
       gnet_tcp_socket_connect_async_cancel (state);
     }
 }
 
 
-void 
+static void 
 gnet_tcp_socket_connect_tcp_cb (GTcpSocket* socket, gpointer data)
 {
   GTcpSocketConnectState* state = (GTcpSocketConnectState*) data;
@@ -180,6 +265,7 @@ gnet_tcp_socket_connect_tcp_cb (GTcpSocket* socket, gpointer data)
       (*state->func)(socket, GTCP_SOCKET_CONNECT_ASYNC_STATUS_OK, state->data);
       state->in_callback = FALSE;
 
+      /* FIXME: don't abuse _cancel() as free function for the state */
       gnet_tcp_socket_connect_async_cancel (state);
 
       return;
@@ -196,9 +282,10 @@ gnet_tcp_socket_connect_tcp_cb (GTcpSocket* socket, gpointer data)
       ia = (GInetAddr*) state->ia_next->data;
       state->ia_next = state->ia_next->next;
 
-      tcp_id = gnet_tcp_socket_new_async (ia, 
-					  gnet_tcp_socket_connect_tcp_cb, 
-					  state);
+      tcp_id = gnet_tcp_socket_new_async_full (ia,
+          gnet_tcp_socket_connect_tcp_cb, state, (GDestroyNotify) NULL,
+          state->context, state->priority);
+
       if (tcp_id)	/* Success */
 	{
 	  state->tcp_id = tcp_id;
@@ -211,6 +298,7 @@ gnet_tcp_socket_connect_tcp_cb (GTcpSocket* socket, gpointer data)
   (*state->func)(NULL, GTCP_SOCKET_CONNECT_ASYNC_STATUS_TCP_ERROR, state->data);
   state->in_callback = FALSE;
 
+  /* FIXME: don't abuse _cancel() as free function for the state */
   gnet_tcp_socket_connect_async_cancel (state);
 }
 
@@ -244,14 +332,15 @@ gnet_tcp_socket_connect_async_cancel (GTcpSocketConnectAsyncID id)
     }
 
   if (state->inetaddr_id)
-    {
-      gnet_inetaddr_new_async_cancel(state->inetaddr_id);
-    }
+      gnet_inetaddr_new_list_async_cancel (state->inetaddr_id);
 
   if (state->tcp_id)
-    {
       gnet_tcp_socket_new_async_cancel (state->tcp_id);
-    }
+
+  if (state->notify)
+    state->notify (state->data);
+
+  g_main_context_unref (state->context);
 
   g_free (state);
 }
@@ -277,7 +366,7 @@ gnet_tcp_socket_new (const GInetAddr* addr)
 
   /* Use SOCKS if enabled */
   if (gnet_socks_get_enabled())
-    return gnet_private_socks_tcp_socket_new (addr);
+    return _gnet_socks_tcp_socket_new (addr);
 
   /* Otherwise, connect directly to the address */
   return gnet_tcp_socket_new_direct (addr);
@@ -299,16 +388,19 @@ gnet_tcp_socket_new (const GInetAddr* addr)
 GTcpSocket* 
 gnet_tcp_socket_new_direct (const GInetAddr* addr)
 {
-  int 			sockfd;
-  GTcpSocket* 		s;
+  SOCKET		sockfd;
+  GTcpSocket* 	s;
   int			rv;
 
   g_return_val_if_fail (addr != NULL, NULL);
 
   /* Create socket */
   sockfd = socket (GNET_INETADDR_FAMILY(addr), SOCK_STREAM, 0);
-  if (sockfd < 0)
-    return NULL;
+  if (!GNET_IS_SOCKET_VALID(sockfd))
+    {
+      g_warning ("socket() failed");
+      return NULL;
+    }
 
   /* Create GTcpSocket */
   s = g_new0 (GTcpSocket, 1);
@@ -332,9 +424,54 @@ gnet_tcp_socket_new_direct (const GInetAddr* addr)
 
 /* **************************************** */
 
+/**
+ *  gnet_tcp_socket_new_async_full:
+ *  @addr: address
+ *  @func: callback function
+ *  @data: data to pass to @func on callback
+ *  @notify: function to call to free @data, or NULL
+ *  @context: the #GMainContext to use for notifications, or NULL for the
+ *      default GLib main context.  If in doubt, pass NULL.
+ *  @priority: the priority with which to schedule notifications in the
+ *      main context, e.g. #G_PRIORITY_DEFAULT or #G_PRIORITY_HIGH.
+ *
+ *  Asynchronously creates a #GTcpSocket and connects to @addr.  The
+ *  callback is called once the connection is made or an error occurs.
+ *  The callback will not be called during the call to this function.
+ *
+ *  SOCKS is used if SOCKS is enabled.  The SOCKS negotiation will
+ *  block.
+ *
+ *  Returns: the ID of the connection; NULL on failure.  The ID can be
+ *  used with gnet_tcp_socket_new_async_cancel() to cancel the
+ *  connection.
+ *
+ *  Since: 2.0.8
+ **/
+GTcpSocketNewAsyncID
+gnet_tcp_socket_new_async_full (const GInetAddr * addr,
+    GTcpSocketNewAsyncFunc func, gpointer data, GDestroyNotify notify,
+    GMainContext * context, gint priority)
+{
+  GTcpSocketNewAsyncID async_id;
+
+  g_return_val_if_fail (addr != NULL, NULL);
+  g_return_val_if_fail (func != NULL, NULL);
+
+  /* Use SOCKS if enabled, otherwise, connect directly to the address */
+  if (gnet_socks_get_enabled()) {
+    async_id = _gnet_socks_tcp_socket_new_async_full (addr, func, data,
+        notify, context, priority);
+  } else {
+    async_id = gnet_tcp_socket_new_async_direct_full (addr, func, data,
+        notify, context, priority);
+  }
+
+  return async_id;
+}
 
 /**
- *  gnet_tcp_socket_new_async
+ *  gnet_tcp_socket_new_async:
  *  @addr: address
  *  @func: callback function
  *  @data: data to pass to @func on callback
@@ -352,24 +489,136 @@ gnet_tcp_socket_new_direct (const GInetAddr* addr)
  *
  **/
 GTcpSocketNewAsyncID
-gnet_tcp_socket_new_async (const GInetAddr* addr, 
-			   GTcpSocketNewAsyncFunc func,
-			   gpointer data)
+gnet_tcp_socket_new_async (const GInetAddr * addr, GTcpSocketNewAsyncFunc func,
+    gpointer data)
 {
   g_return_val_if_fail (addr != NULL, NULL);
   g_return_val_if_fail (func != NULL, NULL);
 
-  /* Use SOCKS if enabled */
-  if (gnet_socks_get_enabled())
-    return gnet_private_socks_tcp_socket_new_async (addr, func, data);
-
-  /* Otherwise, connect directly to the address */
-  return gnet_tcp_socket_new_async_direct (addr, func, data);
+  return gnet_tcp_socket_new_async_full (addr, func, data,
+      (GDestroyNotify) NULL, NULL, G_PRIORITY_DEFAULT);
 }
 
 
-#ifndef GNET_WIN32  /*********** Unix code ***********/
+/**
+ *  gnet_tcp_socket_new_async_direct_full:
+ *  @addr: address
+ *  @func: callback function
+ *  @data: data to pass to @func on callback
+ *  @notify: function to call to free @data, or NULL
+ *  @context: the #GMainContext to use for notifications, or NULL for the
+ *      default GLib main context.  If in doubt, pass NULL.
+ *  @priority: the priority with which to schedule notifications in the
+ *      main context, e.g. #G_PRIORITY_DEFAULT or #G_PRIORITY_HIGH.
+ *
+ *  Asynchronously creates a #GTcpSocket and connects to @addr without
+ *  using SOCKS.  Most users should use gnet_tcp_socket_new_async()
+ *  instead.  The callback is called once the connection is made or an
+ *  error occurs.  The callback will not be called during the call to
+ *  this function.
+ *
+ *  Returns: the ID of the connection; NULL on failure.  The ID can be
+ *  used with gnet_tcp_socket_new_async_cancel() to cancel the
+ *  connection.
+ *
+ *  Since: 2.0.8
+ **/
+GTcpSocketNewAsyncID
+gnet_tcp_socket_new_async_direct_full (const GInetAddr * addr, 
+    GTcpSocketNewAsyncFunc func, gpointer data, GDestroyNotify notify,
+    GMainContext * context, gint priority)
+{
+  SOCKET		sockfd;
+#ifndef GNET_WIN32
+  gint 			flags;
+#else
+  u_long		arg;
+#endif
+  GTcpSocket* 		s;
+  GTcpSocketAsyncState* state;
+  gint			status;
 
+  g_return_val_if_fail (addr != NULL, NULL);
+  g_return_val_if_fail (func != NULL, NULL);
+
+  if (context == NULL)
+    context = g_main_context_default ();
+
+  /* Create socket */
+  sockfd = socket(GNET_INETADDR_FAMILY(addr), SOCK_STREAM, 0);
+  if (!GNET_IS_SOCKET_VALID(sockfd))
+    {
+      g_warning ("socket() failed");
+      return NULL;
+    }
+
+  /* Force the socket into non-blocking mode */
+#ifndef GNET_WIN32
+  /* Get the flags (should all be 0?) */
+  flags = fcntl(sockfd, F_GETFL, 0);
+  if (flags == -1) 
+    {
+      g_warning ("fcntl() failed");
+      GNET_CLOSE_SOCKET(sockfd);    
+      return NULL;
+    }
+
+  if (fcntl(sockfd, F_SETFL, flags | O_NONBLOCK) == -1)
+    {
+      g_warning ("fcntl() failed");
+      GNET_CLOSE_SOCKET(sockfd);    
+      return NULL;
+    }
+#else
+  arg = 1;
+  ioctlsocket(sockfd, FIONBIO, &arg);
+#endif
+
+  /* Create our structure */
+  s = g_new0(GTcpSocket, 1);
+  s->ref_count = 1;
+  s->sockfd = sockfd;
+
+  /* Connect (but non-blocking!) */
+  status = connect(s->sockfd, &GNET_INETADDR_SA(addr), 
+		   GNET_INETADDR_LEN(addr));
+  if (STATUS_IS_SOCKET_ERROR (status))
+    {
+      if (!ERROR_IS_CONNECT_IN_PROGRESS ())
+	{
+	  GNET_CLOSE_SOCKET(sockfd);    
+	  g_free(s);
+	  return NULL;
+	}
+    }
+
+  /* Save address */ 
+  s->sa = addr->sa;
+
+  /* Note that if connect returns 0, then we're already connected and
+     we could call the call back immediately.  But, it would probably
+     make things too complicated for the user if we could call the
+     callback before we returned from this function.  */
+
+  /* Wait for the connection */
+  state = g_new0 (GTcpSocketAsyncState, 1);
+  state->socket = s;
+  state->func = func;
+  state->data = data;
+  state->notify = notify;
+#ifndef GNET_WIN32
+  state->flags = flags;
+#endif
+  state->iochannel = g_io_channel_ref (gnet_tcp_socket_get_io_channel (s));
+  state->context = g_main_context_ref (context);
+  state->priority = priority;
+
+  state->connect_watch = _gnet_io_watch_add_full (state->context,
+      state->priority, state->iochannel, GNET_ANY_IO_CONDITION,
+      gnet_tcp_socket_new_async_cb, state, NULL);
+
+  return state;
+}
 
 /**
  *  gnet_tcp_socket_new_async_direct
@@ -389,85 +638,30 @@ gnet_tcp_socket_new_async (const GInetAddr* addr,
  *
  **/
 GTcpSocketNewAsyncID
-gnet_tcp_socket_new_async_direct (const GInetAddr* addr, 
-				  GTcpSocketNewAsyncFunc func,
-				  gpointer data)
+gnet_tcp_socket_new_async_direct (const GInetAddr * addr, 
+    GTcpSocketNewAsyncFunc func, gpointer data)
 {
-  gint 			sockfd;
-  gint 			flags;
-  GTcpSocket* 		s;
-  GTcpSocketAsyncState* state;
+  GTcpSocketNewAsyncID async_id;
 
-  g_return_val_if_fail(addr != NULL, NULL);
-  g_return_val_if_fail(func != NULL, NULL);
+  g_return_val_if_fail (addr != NULL, NULL);
+  g_return_val_if_fail (func != NULL, NULL);
 
-  /* Create socket */
-  sockfd = socket(GNET_INETADDR_FAMILY(addr), SOCK_STREAM, 0);
-  if (sockfd < 0)
-    return NULL;
+  async_id = gnet_tcp_socket_new_async_direct_full (addr, func, data,
+      (GDestroyNotify) NULL, NULL, G_PRIORITY_DEFAULT);
 
-  /* Get the flags (should all be 0?) */
-  flags = fcntl(sockfd, F_GETFL, 0);
-  if (flags == -1)
-    return NULL;
-
-  if (fcntl(sockfd, F_SETFL, flags | O_NONBLOCK) == -1)
-    return NULL;
-
-  /* Create our structure */
-  s = g_new0(GTcpSocket, 1);
-  s->ref_count = 1;
-  s->sockfd = sockfd;
-
-  /* Connect (but non-blocking!) */
-  if (connect(s->sockfd, &GNET_INETADDR_SA(addr), 
-	      GNET_INETADDR_LEN(addr)) < 0)
-    {
-      if (errno != EINPROGRESS)
-	{
-	  g_free(s);
-	  return NULL;
-	}
-    }
-
-  /* Save address */ 
-  s->sa = addr->sa;
-
-  /* Note that if connect returns 0, then we're already connected and
-     we could call the call back immediately.  But, it would probably
-     make things too complicated for the user if we could call the
-     callback before we returned from this function.  */
-
-  /* Wait for the connection */
-  state = g_new0(GTcpSocketAsyncState, 1);
-  state->socket = s;
-  state->func = func;
-  state->data = data;
-  state->flags = flags;
-  state->iochannel = gnet_private_io_channel_new(s->sockfd);
-  state->connect_watch = g_io_add_watch(state->iochannel,
-					GNET_ANY_IO_CONDITION,
-					gnet_tcp_socket_new_async_cb, 
-					state);
-
-  return state;
+  return async_id;
 }
 
 
-gboolean 
+static gboolean 
 gnet_tcp_socket_new_async_cb (GIOChannel* iochannel, 
 			      GIOCondition condition, 
 			      gpointer data)
 {
   GTcpSocketAsyncState* state = (GTcpSocketAsyncState*) data;
-  gint error, len;
+  socklen_t len;
+  gint error;
 
-  g_source_remove (state->connect_watch);
-  state->connect_watch = 0;
-  g_io_channel_unref (state->iochannel);
-  state->iochannel = NULL;
-
-  errno = 0;
   if (!((condition & G_IO_IN) || (condition & G_IO_OUT)))
     goto error;
 
@@ -475,28 +669,46 @@ gnet_tcp_socket_new_async_cb (GIOChannel* iochannel,
 
   /* Get the error option */
   if (getsockopt(state->socket->sockfd, SOL_SOCKET, SO_ERROR, (void*) &error, &len) < 0)
-    goto error;
+    {
+      g_warning ("getsockopt() failed");
+      goto error;
+    }
 
   /* Check if there is an error */
   if (error)
     goto error;
 
+#ifndef GNET_WIN32
   /* Reset the flags */
   if (fcntl(state->socket->sockfd, F_SETFL, state->flags) != 0)
-    goto error;
+    {
+      g_warning ("fcntl() failed");
+      goto error;
+    }
+#endif
 
   /* Success */
-  (*state->func)(state->socket, state->data);
-  g_free(state);
-  return FALSE;
+  (*state->func) (state->socket, state->data);
 
-  /* Error */
- error:
-  (*state->func)(NULL, state->data);
-  gnet_tcp_socket_delete (state->socket);
-  g_free(state);
+done:
 
-  return FALSE;
+  state->connect_watch = 0;
+  g_io_channel_unref (state->iochannel);
+  g_main_context_unref (state->context);
+  /* FIXME: is notify really useful in this context? */
+  if (state->notify)
+    state->notify (state->data);
+  memset (state, 0xaa, sizeof (*state));
+  g_free (state);
+  return FALSE; /* don't call us again */
+
+/* ERROR */
+error:
+  {
+    (*state->func) (NULL, state->data);
+    gnet_tcp_socket_delete (state->socket);
+    goto done;
+  }
 }
 
 
@@ -514,114 +726,16 @@ gnet_tcp_socket_new_async_cancel (GTcpSocketNewAsyncID id)
   GTcpSocketAsyncState* state = (GTcpSocketAsyncState*) id;
 
   if (state->connect_watch)
-    g_source_remove(state->connect_watch);
+    _gnet_source_remove (state->context, state->connect_watch);
   if (state->iochannel)
     g_io_channel_unref (state->iochannel);
   gnet_tcp_socket_delete (state->socket);
+  g_main_context_unref (state->context);
+  if (state->notify)
+    state->notify (state->data);
   g_free (state);
 }
 
-#else	/*********** Windows code ***********/
-
-
-GTcpSocketNewAsyncID
-gnet_tcp_socket_new_async_direct (const GInetAddr* addr,
-				  GTcpSocketNewAsyncFunc func,
-				  gpointer data)
-{
-  gint sockfd;
-  gint status;
-  GTcpSocket* s;
-  GTcpSocketAsyncState* state;
-  u_long arg;
-
-  g_return_val_if_fail(addr != NULL, NULL);
-  g_return_val_if_fail(func != NULL, NULL);
-
-  /* Create socket */
-  sockfd = socket(GNET_INETADDR_FAMILY(addr), SOCK_STREAM, 0);
-  if (sockfd == INVALID_SOCKET)
-      return NULL;
-	
-  /* Create our structure */
-  s = g_new0(GTcpSocket, 1);
-  s->ref_count = 1;
-  s->sockfd = sockfd;
-
-  /* Force the socket into non-blocking mode */
-  arg = 1;
-  ioctlsocket(sockfd, FIONBIO, &arg);
-
-  status = connect(s->sockfd, &GNET_INETADDR_SA(addr), 
-		   GNET_INETADDR_LEN(addr));
-  if (status == SOCKET_ERROR) /* Returning an error is ok, unless.. */
-    {
-      status = WSAGetLastError();
-      if (status != WSAEWOULDBLOCK)
-	{
-	  g_free(s);
-	  return NULL;
-	}
-    }
-
-  /* Save address */ 
-  s->sa = addr->sa;
-
-  /* Wait for the connection */
-  state = g_new0(GTcpSocketAsyncState, 1);
-  state->socket = s;
-  state->func = func;
-  state->data = data;
-  state->socket->sockfd = sockfd;
-
-  state->connect_watch = 
-    g_io_add_watch(gnet_private_io_channel_new(s->sockfd),
-		   G_IO_IN | G_IO_ERR,
-		   gnet_tcp_socket_new_async_cb, 
-		   state);
-
-  if (state->connect_watch <= 0)
-      return NULL;
-
-  return state;
-}
-
-
-
-gboolean
-gnet_tcp_socket_new_async_cb (GIOChannel* iochannel,
-			      GIOCondition condition,
-			      gpointer data)
-{
-  GTcpSocketAsyncState* state = (GTcpSocketAsyncState*) data;
-
-  if (condition & G_IO_ERR)
-    goto error;
-
-  (*state->func)(state->socket, state->data);
-  g_free (state);
-  return FALSE;
-
- error:
-  (*state->func)(NULL, state->data);
-  gnet_tcp_socket_delete (state->socket);
-  g_free (state);
-  return FALSE;
-}
-
-
-void
-gnet_tcp_socket_new_async_cancel (GTcpSocketNewAsyncID id)
-{
-  GTcpSocketAsyncState* state = (GTcpSocketAsyncState*) id;
-	
-  g_source_remove(state->connect_watch);
-  gnet_tcp_socket_delete(state->socket);
-  g_free (state);
-}
-
-#endif		/*********** End Windows code ***********/
-
 
 
 /**
@@ -652,9 +766,27 @@ gnet_tcp_socket_ref (GTcpSocket* socket)
 {
   g_return_if_fail (socket != NULL);
 
-  ++socket->ref_count;
+  g_atomic_int_inc (&socket->ref_count);
 }
 
+/* returns TRUE if the socket has been deleted (useful for io watches) */
+static gboolean
+gnet_tcp_socket_unref_internal (GTcpSocket * socket)
+{
+  if (!g_atomic_int_dec_and_test (&socket->ref_count))
+    return FALSE;
+
+  if (socket->accept_watch)
+    g_source_remove (socket->accept_watch);
+
+  GNET_CLOSE_SOCKET (socket->sockfd); /* Don't care if this fails... */
+
+  if (socket->iochannel)
+    g_io_channel_unref (socket->iochannel);
+
+  g_free (socket);
+  return TRUE;
+}
 
 /**
  *  gnet_tcp_socket_unref
@@ -669,20 +801,7 @@ gnet_tcp_socket_unref (GTcpSocket* socket)
 {
   g_return_if_fail (socket != NULL);
 
-  socket->ref_count--;
-
-  if (socket->ref_count == 0)
-    {
-      if (socket->accept_watch)
-	g_source_remove (socket->accept_watch);
-
-      GNET_CLOSE_SOCKET(socket->sockfd);/* Don't care if this fails... */
-
-      if (socket->iochannel)
-	g_io_channel_unref (socket->iochannel);
-
-      g_free(socket);
-    }
+  gnet_tcp_socket_unref_internal (socket);
 }
 
 
@@ -716,7 +835,7 @@ gnet_tcp_socket_get_io_channel (GTcpSocket* socket)
   g_return_val_if_fail (socket != NULL, NULL);
 
   if (socket->iochannel == NULL)
-    socket->iochannel = gnet_private_io_channel_new(socket->sockfd);
+    socket->iochannel = _gnet_io_channel_new(socket->sockfd);
 
   return socket->iochannel;
 }
@@ -753,7 +872,8 @@ gnet_tcp_socket_get_remote_inetaddr (const GTcpSocket* socket)
  *
  *  Gets the local host's address from a #GTcpSocket.
  *
- *  Returns: a #GInetAddr.
+ *  Returns: a #GInetAddr, or NULL on error.  Unref with gnet_inetaddr_unref()
+ *  when no longer needed.
  *
  **/
 GInetAddr*  
@@ -904,40 +1024,56 @@ gnet_tcp_socket_server_new_with_port (gint port)
 GTcpSocket* 
 gnet_tcp_socket_server_new_full (const GInetAddr* iface, gint port)
 {
-  int sockfd = 0;
+  SOCKET sockfd = 0;
   struct sockaddr_storage sa;
   GTcpSocket* s;
   socklen_t socklen;
+  const int on = 1;
+#ifndef GNET_WIN32
+  gint flags;
+#endif
+
 
   /* Use SOCKS if enabled */
   if (!iface && gnet_socks_get_enabled())
-    return gnet_private_socks_tcp_socket_server_new (port);
+    return _gnet_socks_tcp_socket_server_new (port);
 
   /* Create sockfd and address */
-  sockfd = gnet_private_create_listen_socket (SOCK_STREAM, iface, port, &sa);
-  if (sockfd < 0)
+  sockfd = _gnet_create_listen_socket (SOCK_STREAM, iface, port, &sa);
+  if (!GNET_IS_SOCKET_VALID(sockfd))
     return NULL;
+  
+  /* Set REUSEADDR so we can reuse the port */
+  if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, 
+	  (void*) &on, sizeof(on)) != 0)
+	      g_warning("Can't set reuse on tcp socket\n");
 
-  /* The socket is set to non-blocking mode later in the Windows
-     version.*/
 #ifndef GNET_WIN32		/* Unix */
   {
-    gint flags;
-    const int on = 1;
-    /* Set REUSEADDR so we can reuse the port */
-    if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, 
-		   (void*) &on, sizeof(on)) != 0)
-      g_warning("Can't set reuse on tcp socket\n");
-
     /* Get the flags (should all be 0?) */
     flags = fcntl(sockfd, F_GETFL, 0);
     if (flags == -1)
-      goto error;
+      {
+	g_warning ("fcntl() failed");
+	goto error;
+      }
     
     /* Make the socket non-blocking */
     if (fcntl(sockfd, F_SETFL, flags | O_NONBLOCK) == -1)
-      goto error;
+      {
+	g_warning ("fcntl() failed");
+	goto error;
+      }
   }
+#else /* Windows */
+  {
+    u_long arg;
+
+	/* Make the socket non-blocking */
+    arg = 1;
+    if (ioctlsocket(sockfd, FIONBIO, &arg))
+      goto error;
+    }
 #endif
 
   /* Bind */
@@ -999,7 +1135,7 @@ gnet_tcp_socket_server_accept (GTcpSocket* socket)
   g_return_val_if_fail (socket != NULL, NULL);
 
   if (gnet_socks_get_enabled())
-    return gnet_private_socks_tcp_socket_server_accept(socket);
+    return _gnet_socks_tcp_socket_server_accept(socket);
 
  try_again:
   
@@ -1071,7 +1207,7 @@ gnet_tcp_socket_server_accept_nonblock (GTcpSocket* socket)
   g_return_val_if_fail (socket != NULL, NULL);
 
   if (gnet_socks_get_enabled())
-    return gnet_private_socks_tcp_socket_server_accept(socket);
+    return _gnet_socks_tcp_socket_server_accept(socket);
 
  try_again:
 
@@ -1111,7 +1247,7 @@ gnet_tcp_socket_server_accept_nonblock (GTcpSocket* socket)
 GTcpSocket*
 gnet_tcp_socket_server_accept (GTcpSocket* socket)
 {
-  gint sockfd;
+  SOCKET sockfd;
   struct sockaddr_storage sa;
   fd_set fdset;
   GTcpSocket* s;
@@ -1120,12 +1256,12 @@ gnet_tcp_socket_server_accept (GTcpSocket* socket)
   g_return_val_if_fail (socket != NULL, NULL);
 
   if (gnet_socks_get_enabled())
-    return gnet_private_socks_tcp_socket_server_accept(socket);
+    return _gnet_socks_tcp_socket_server_accept(socket);
 	
   FD_ZERO(&fdset);
   FD_SET((unsigned)socket->sockfd, &fdset);
 
-  if (select(socket->sockfd + 1, &fdset, NULL, NULL, NULL) == -1)
+  if (select((int)socket->sockfd + 1, &fdset, NULL, NULL, NULL) == -1)
       return NULL;
 
   /* Don't force the socket into blocking mode */
@@ -1134,7 +1270,7 @@ gnet_tcp_socket_server_accept (GTcpSocket* socket)
   sockfd = accept(socket->sockfd, (struct sockaddr*) &sa, &n);
   /* if it fails, looping isn't going to help */
 
-  if (sockfd == INVALID_SOCKET)
+  if (!GNET_IS_SOCKET_VALID(sockfd))
       return NULL;
 
   s = g_new0(GTcpSocket, 1);
@@ -1149,7 +1285,7 @@ gnet_tcp_socket_server_accept (GTcpSocket* socket)
 GTcpSocket*
 gnet_tcp_socket_server_accept_nonblock (GTcpSocket* socket)
 {
-  gint sockfd;
+  SOCKET sockfd;
   struct sockaddr_storage sa;
 
   fd_set fdset;
@@ -1160,19 +1296,17 @@ gnet_tcp_socket_server_accept_nonblock (GTcpSocket* socket)
   g_return_val_if_fail (socket != NULL, NULL);
 
   if (gnet_socks_get_enabled())
-    return gnet_private_socks_tcp_socket_server_accept(socket);
+    return _gnet_socks_tcp_socket_server_accept(socket);
 
   FD_ZERO(&fdset);
   FD_SET((unsigned)socket->sockfd, &fdset);
 
-  if (select(socket->sockfd + 1, &fdset, NULL, NULL, NULL) == -1)
-    {
+  if (select((int)socket->sockfd + 1, &fdset, NULL, NULL, NULL) == -1)
       return NULL;
-    }
-  /* make sure the socket is in non-blocking mode */
 
+  /* make sure the socket is in non-blocking mode */
   arg = 1;
-  if(ioctlsocket(socket->sockfd, FIONBIO, &arg))
+  if (ioctlsocket(socket->sockfd, FIONBIO, &arg))
     return NULL;
 
   n = sizeof(sa);
@@ -1180,9 +1314,7 @@ gnet_tcp_socket_server_accept_nonblock (GTcpSocket* socket)
   /* if it fails, looping isn't going to help */
 
   if (sockfd == INVALID_SOCKET)
-    {
       return NULL;
-    }
 
   s = g_new0(GTcpSocket, 1);
   s->ref_count = 1;
@@ -1226,7 +1358,7 @@ gnet_tcp_socket_server_accept_async (GTcpSocket* socket,
 
   if (gnet_socks_get_enabled())
     {
-      gnet_private_socks_tcp_socket_server_accept_async (socket, accept_func, user_data);
+      _gnet_socks_tcp_socket_server_accept_async (socket, accept_func, user_data);
       return;
     }
 
@@ -1253,7 +1385,6 @@ tcp_socket_server_accept_async_cb (GIOChannel* iochannel, GIOCondition condition
   if (condition & G_IO_IN)
     {
       GTcpSocket* client;
-      gboolean destroyed = FALSE;
 
       client = gnet_tcp_socket_server_accept_nonblock (server);
       if (!client) 
@@ -1264,10 +1395,7 @@ tcp_socket_server_accept_async_cb (GIOChannel* iochannel, GIOCondition condition
 
       (server->accept_func)(server, client, server->accept_data);
 
-      if (server->ref_count == 1)
-	destroyed = TRUE;
-      gnet_tcp_socket_unref (server);
-      if (destroyed || !server->accept_watch)
+      if (gnet_tcp_socket_unref_internal (server) || !server->accept_watch)
 	return FALSE;
     }
   else
