@@ -1,5 +1,6 @@
-/* GNet - Networking library
+/* GNet - Networking library URI parser
  * Copyright (C) 2000-2003  David Helder, David Bolcsfoldi, Eric Williams
+ * Copyright (C) 2007       Tim-Philipp Müller <tim centricular net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -21,14 +22,14 @@
 #include "uri.h"
 
 static void   field_unescape (gchar *str);
-static gchar* field_escape (gchar* str, guchar mask);
+static gchar* field_escape (gchar *str, guchar mask);
 
 #define USERINFO_ESCAPE_MASK	0x01
 #define PATH_ESCAPE_MASK	0x02
 #define QUERY_ESCAPE_MASK	0x04
 #define FRAGMENT_ESCAPE_MASK	0x08
 
-static guchar neednt_escape_table[] = 
+static const guchar neednt_escape_table[] = 
 {
 	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
 	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
@@ -99,11 +100,221 @@ for ($i = 0; $i < 32; $i++)
 }
 */
 
+typedef struct {
+  const gchar *str;
+  guint len;
+} StringPart;
 
-/* our own ISSPACE.  ANSI isspace is local dependent */
+#define DUP_STRING_PART(sp)  (sp.len > 0) ? g_strndup (sp.str, sp.len) : NULL
+
+/* our own ISSPACE.  ANSI isspace is locale dependent and g_ascii_isspace()
+ * does not recognise #10 as space */
 #define ISSPACE(C) (((C) >= 9 && (C) <= 13) || (C) == ' ')
 
+static gboolean
+gnet_uri_parse (const gchar * uri, StringPart * scheme, StringPart * userinfo,
+    StringPart * hostname, guint * port, StringPart * path,
+    StringPart * query, StringPart * fragment)
+{
+  const StringPart empty = { NULL, 0 };
+  const gchar *p, *temp;
 
+  *scheme = *userinfo = *hostname = *path = *query = *fragment = empty;
+  *port = 0;
+
+  /* Skip initial whitespaces */
+  p = uri;
+  while (*p && ISSPACE(*p))
+    ++p;
+
+  /* Error out if it's just a string of space */
+  if (*p == '\0')
+    return FALSE;
+
+  /* Scheme */
+  temp = p;
+  while (*p && *p != ':' && *p != '/' && *p != '?' && *p != '#')
+    ++p;
+  if (*p == ':') {
+    scheme->str = temp;
+    scheme->len = (guint) (p - temp);
+    ++p;
+  } else {
+    /* This char is NUL, /, ?, or # */
+    p = temp;
+  }
+
+  /* Authority */
+  if (*p == '/' && p[1] == '/') {
+    p += 2;
+
+    /* Userinfo */
+    temp = p;
+    while (*p && *p != '@' && *p != '/' ) /* Look for @ or / */
+      ++p;
+    if (*p == '@') { /* Found userinfo */
+      userinfo->str = temp;
+      userinfo->len = (guint) (p - temp);
+      ++p;
+    } else {        
+      p = temp;
+    }
+
+    /* Hostname */
+
+    /* Check for no hostname at all (e.g. file:// URIs) */
+    if (*p == '/')
+      goto path;
+
+    /* Check for IPv6 canonical hostname in brackets */
+    if (*p == '[') {
+      p++;  /* Skip [ */
+      temp = p;
+
+      while (*p && *p != ']')
+        ++p;
+
+      if ((p - temp) == 0)
+        return FALSE;
+
+      hostname->str = temp;
+      hostname->len = (guint) (p - temp);
+
+      if (*p)
+        ++p;  /* Skip ] (if there) */
+    } else {
+      temp = p;
+      while (*p && *p != '/' && *p != '?' && *p != '#' && *p != ':')
+        ++p;
+
+      if ((p - temp) == 0) 
+        return FALSE;
+
+      hostname->str = temp;
+      hostname->len = (guint) (p - temp);
+    }
+
+    /* Port */
+    if (*p == ':') {
+      for (++p; isdigit((int)*p); ++p)
+        *port = (*port) * 10 + (*p - '0');
+    }
+  }
+
+  /* Path (we are liberal and won't check if it starts with /) */
+
+path:
+
+  temp = p;
+  while (*p && *p != '?' && *p != '#')
+    ++p;
+  if (p != temp) {
+    path->str = temp;
+    path->len = (guint) (p - temp);
+  }
+
+  /* Query */
+  if (*p == '?') {
+    temp = p + 1;
+    while (*p && *p != '#')
+      ++p;
+    query->str = temp;
+    query->len = (guint) (p - temp);
+  }
+
+  /* Fragment */
+  if (*p == '#') {
+    ++p;
+    fragment->str = p;
+    fragment->len = strlen (p);
+  }
+
+  return TRUE;
+}
+
+static gchar *
+parse_inplace_munge_string_part (StringPart * part)
+{
+  if (part->len == 0)
+    return NULL;
+  ((gchar*)part->str)[part->len] = '\0';
+  return (gchar *) part->str;
+}
+
+/**
+ *  gnet_uri_parse_inplace:
+ *  @guri: pointer to an uninitialised #GURI structure (usually on the stack)
+ *  @uri: a writable URI string (which will be modified by this function!)
+ *  @hostname: a preallocated array of size @len, into which the hostname will
+ *             be copied if the URI has a hostname. This array should live
+ *             as long as @guri may be used, since @guri might reference it.
+ *             You may pass NULL here if you know your URIs won't contain a
+ *             hostname field
+ *  @size: size of @hostname in bytes
+ *
+ *  This function parses an URI string very efficiently and without allocating
+ *  any memory on the heap. It does this by modifying the passed-in string
+ *  @uri and changing field separators into string terminators in place. The
+ *  result will then be put into the already-allocated and usually
+ *  uninitialised, GURI structure @guri.
+ *
+ *  Because the slash (') which separates the hostname from the path in an URI
+ *  is needed as the first character of the path, it cannot be turned into a
+ *  string terminator for the hostname. This makes it necessary for the caller
+ *  (ie. you) to pass in an array where the hostname can be put.
+ *
+ *  If the URI has been parsed successfully, the fields of the GURI structure
+ *  will either point into the given string @uri (or to @hostname in case of
+ *  the hostname field), or be NULL. No freeing or de-initialising of the
+ *  GURI structure is required.
+ *
+ *  The only gnet_uri_*() functions you are allowed to call with this URI
+ *  structure are gnet_uri_clone(), gnet_uri_equal(), gnet_uri_hash(),
+ *  gnet_uri_get_string() and gnet_uri_unescape() (the latter unescapes the
+ *  string fragments in place, so is safe to use here).
+ *
+ *  You must not call any other gnet_uri_*() functions with this GURI,
+ *  especially not any gnet_uri_set() functions.
+ *
+ *  Under normal circumstances, you should never need to use this function.
+ *  Use gnet_uri_new() instead.
+ *  
+ *  Returns: TRUE if @uri was parsed successfully and #GURI is usable,
+ *           or FALSE if there was a failure.
+ *
+ *  Since: 2.0.8
+ *
+ **/
+gboolean
+gnet_uri_parse_inplace (GURI * guri, gchar * uri, gchar * hostname, gsize len)
+{
+  StringPart scheme, usr, host, path, query, frag;
+  guint port;
+
+  if (!gnet_uri_parse (uri, &scheme, &usr, &host, &port, &path, &query, &frag))
+    return FALSE;
+
+  if (host.len >= len)
+    return FALSE;
+
+  if (host.len > 0) {
+    if (hostname == NULL)
+      return FALSE;
+
+    strncpy (hostname, host.str, host.len);
+    hostname[host.len] = '\0';
+  }
+
+  guri->scheme = parse_inplace_munge_string_part (&scheme);
+  guri->userinfo = parse_inplace_munge_string_part (&usr);
+  guri->hostname = (host.len > 0) ? hostname : NULL;
+  guri->path = parse_inplace_munge_string_part (&path);
+  guri->query = parse_inplace_munge_string_part (&query);
+  guri->fragment = parse_inplace_munge_string_part (&frag);
+  guri->port = port;
+
+  return TRUE;
+}
 
 /**
  *  gnet_uri_new
@@ -121,113 +332,28 @@ for ($i = 0; $i < 32; $i++)
  *  Returns: a new #GURI, or NULL if there was a failure.
  *
  **/
-GURI* 
-gnet_uri_new (const gchar* uri)
+GURI *
+gnet_uri_new (const gchar * uri)
 {
-  GURI* guri = NULL;
-  const gchar* p;
-  const gchar* temp;
+  StringPart scheme, usr, host, path, query, frag;
+  guint port;
+  GURI *guri;
 
-  g_return_val_if_fail (uri, NULL);
+  g_return_val_if_fail (uri != NULL, NULL);
 
-  /* Skip initial whitespace */
-  p = uri;
-  while (*p && ISSPACE((int)*p))
-    ++p;
-  if (!*p)	/* Error if it's just a string of space */
+  if (!gnet_uri_parse (uri, &scheme, &usr, &host, &port, &path, &query, &frag))
     return NULL;
 
-  guri = g_new0 (GURI, 1);
-
-  /* Scheme */
-  temp = p;
-  while (*p && *p != ':' && *p != '/' && *p != '?' && *p != '#')
-    ++p;
-  if (*p == ':')
-    {
-      guri->scheme = g_strndup (temp, p - temp);
-      ++p;
-    }
-  else	/* This char is NUL, /, ?, or # */
-    p = temp;
-
-  /* Authority */
-  if (*p == '/' && p[1] == '/')
-    {
-      p += 2;
-
-      /* Userinfo */
-      temp = p;
-      while (*p && *p != '@' && *p != '/' ) /* Look for @ or / */
-	++p;
-      if (*p == '@') /* Found userinfo */
-	{
-	  guri->userinfo = g_strndup (temp, p - temp);
-	  ++p;
-	}
-      else
-	p = temp;
-
-      /* Hostname */
-
-      /* Check for IPv6 canonical hostname in brackets */
-      if (*p == '[')
-	{
-	  p++;  /* Skip [ */
-	  temp = p;
-	  while (*p && *p != ']') ++p;
-	  if ((p - temp) == 0)
-	    goto error;
-	  guri->hostname = g_strndup (temp, p - temp);
-	  if (*p)
-	    p++;	/* Skip ] (if there) */
-	}
-      else
-	{
-	  temp = p;
-	  while (*p && *p != '/' && *p != '?' && *p != '#' && *p != ':') ++p;
-	  if ((p - temp) == 0) 
-	    goto error;
-	  guri->hostname = g_strndup (temp, p - temp);
-	}
-
-      /* Port */
-      if (*p == ':')
-	{
-	  for (++p; isdigit((int)*p); ++p)
-	    guri->port = guri->port * 10 + (*p - '0');
-	}
-
-    }
-
-  /* Path (we are liberal and won't check if it starts with /) */
-  temp = p;
-  while (*p && *p != '?' && *p != '#')
-    ++p;
-  if (p != temp)
-    guri->path = g_strndup(temp, p - temp);
-
-  /* Query */
-  if (*p == '?')
-    {
-      temp = p + 1;
-      while (*p && *p != '#')
-        ++p;
-      guri->query = g_strndup (temp, p - temp);
-    }
-
-  /* Fragment */
-  if (*p == '#')
-    {
-      ++p;
-      guri->fragment = g_strdup (p);
-    }
+  guri = g_new (GURI, 1);
+  guri->scheme = DUP_STRING_PART (scheme);
+  guri->userinfo = DUP_STRING_PART (usr);
+  guri->hostname = DUP_STRING_PART (host);
+  guri->path = DUP_STRING_PART (path);
+  guri->query = DUP_STRING_PART (query);
+  guri->fragment = DUP_STRING_PART (frag);
+  guri->port = port;
 
   return guri;
-
- error:
-  gnet_uri_delete (guri);
-  return NULL;
 }
 
 
@@ -459,14 +585,17 @@ gnet_uri_unescape (GURI* uri)
 
 
 static gchar*
-field_escape (gchar* str, guchar mask)
+field_escape (gchar *signed_str, guchar mask)
 {
   gint len;
   gint i;
   gboolean must_escape = FALSE;
-  gchar* dst;
+  guchar *str;
+  gchar *dst;
   gint j;
 
+  str = (guchar*) signed_str;
+
   if (str == NULL)
     return NULL;
 
@@ -474,7 +603,7 @@ field_escape (gchar* str, guchar mask)
   len = 0;
   for (i = 0; str[i]; i++)
     {
-      if (neednt_escape_table[(guint) str[i]] & mask)
+      if (neednt_escape_table[str[i]] & mask)
 	len++;
       else
 	{
@@ -485,7 +614,7 @@ field_escape (gchar* str, guchar mask)
 
   /* Don't escape if unnecessary */
   if (must_escape == FALSE)
-    return str;
+    return signed_str;
 	
   /* Allocate buffer */
   dst = (gchar*) g_malloc(len + 1);
@@ -494,7 +623,7 @@ field_escape (gchar* str, guchar mask)
   for (i = j = 0; str[i]; i++, j++)
     {
       /* Unescaped character */
-      if (neednt_escape_table[(guint) str[i]] & mask)
+      if (neednt_escape_table[str[i]] & mask)
 	{
 	  dst[j] = str[i];
 	}
@@ -504,10 +633,10 @@ field_escape (gchar* str, guchar mask)
 	{
 	  dst[j] = '%';
 
-	  if ((str[i] >> 4) < 10)
-	    dst[j+1] = (str[i] >> 4) + '0';
+	  if (((str[i] & 0xf0) >> 4) < 10)
+	    dst[j+1] = ((str[i] & 0xf0) >> 4) + '0';
 	  else
-	    dst[j+1] = (str[i] >> 4) + 'a' - 10;
+	    dst[j+1] = ((str[i] & 0xf0) >> 4) + 'a' - 10;
 
 	  if ((str[i] & 0x0f) < 10)
 	    dst[j+2] = (str[i] & 0x0f) + '0';
@@ -519,7 +648,7 @@ field_escape (gchar* str, guchar mask)
     }
   dst[j] = '\0';
 
-  g_free (str);
+  g_free (signed_str);
   return dst;
 }
 
@@ -561,7 +690,11 @@ field_unescape (gchar* s)
       else
 	{
 	regular_copy:
-	  *dst = *src;
+	  /* micro-optimisation: a lot of URIs do not include escape
+	   * sequences. By testing the pointer addresses we can avoid
+	   * a lot of reading+writing of the same data */
+	  if (dst != src)
+	    *dst = *src;
 	}
     }
 
@@ -594,8 +727,11 @@ gnet_uri_get_string (const GURI* uri)
   if (uri->scheme)
     g_string_sprintfa (buffer, "%s:", uri->scheme);
 
-  if (uri->userinfo || uri->hostname || uri->port)
-    g_string_append (buffer, "//");
+  if (uri->userinfo  || uri->hostname  || uri->port ||
+      (uri->scheme && uri->hostname == NULL && uri->path && *uri->path == '/'))
+    {
+      g_string_append (buffer, "//");
+    }
 
   if (uri->userinfo)
     {
@@ -650,11 +786,8 @@ gnet_uri_set_scheme (GURI* uri, const gchar* scheme)
 {
   g_return_if_fail (uri);
 
-  if (uri->scheme)
-    {
-      g_free (uri->scheme);
-      uri->scheme = NULL;
-    }
+  g_free (uri->scheme);
+  uri->scheme = NULL;
 
   if (scheme)
     uri->scheme = g_strdup (scheme);
@@ -674,11 +807,8 @@ gnet_uri_set_userinfo (GURI* uri, const gchar* userinfo)
 {
   g_return_if_fail (uri);
 
-  if (uri->userinfo)
-    {
-      g_free (uri->userinfo);
-      uri->userinfo = NULL;
-    }
+  g_free (uri->userinfo);
+  uri->userinfo = NULL;
 
   if (userinfo)
     uri->userinfo = g_strdup (userinfo);
@@ -698,11 +828,8 @@ gnet_uri_set_hostname (GURI* uri, const gchar* hostname)
 {
   g_return_if_fail (uri);
 
-  if (uri->hostname)
-    {
-      g_free (uri->hostname);
-      uri->hostname = NULL;
-    }
+  g_free (uri->hostname);
+  uri->hostname = NULL;
 
   if (hostname)
     uri->hostname = g_strdup (hostname);
@@ -737,11 +864,8 @@ gnet_uri_set_path (GURI* uri, const gchar* path)
 {
   g_return_if_fail (uri);
 
-  if (uri->path)
-    {
-      g_free (uri->path);
-      uri->path = NULL;
-    }
+  g_free (uri->path);
+  uri->path = NULL;
 
   if (path)
     uri->path = g_strdup (path);
@@ -762,11 +886,8 @@ gnet_uri_set_query (GURI* uri, const gchar* query)
 {
   g_return_if_fail (uri);
 
-  if (uri->query)
-    {
-      g_free (uri->query);
-      uri->query = NULL;
-    }
+  g_free (uri->query);
+  uri->query = NULL;
 
   if (query)
     uri->query = g_strdup (query);
@@ -786,11 +907,8 @@ gnet_uri_set_fragment (GURI* uri, const gchar* fragment)
 {
   g_return_if_fail (uri);
 
-  if (uri->fragment)
-    {
-      g_free (uri->fragment);
-      uri->fragment = NULL;
-    }
+  g_free (uri->fragment);
+  uri->fragment = NULL;
 
   if (fragment)
     uri->fragment = g_strdup (fragment);
